#pragma once

#define INTC_BASE 0x48200000U

#define INTC_NUM_INTERRUPTS 128U

#define INTC_SIR_IRQ_OFFSET 0x40U
#define INTC_SIR_FIQ_OFFSET 0x44U
#define INTC_SIR_ACTIVE_MASK 0x7FU

#define INTC_CONTROL_OFFSET 0x48U
#define INTC_CONTROL_NEWIRQAGR 0x1U
#define INTC_CONTROL_NEWFIQAGR 0x2U

#define INTC_THRESHOLD_OFFSET 0x68U

#define INTC_PRIORITY_MIN 0x3FU
#define INTC_THRESHOLD_DISABLED 0xFFU

#define INTC_ISR_CLEAR_OFFSET(g) (0x94U + (g) * 32U)

#if !defined(__ASSEMBLER__)

#include <rt/assert.h>

#include <stddef.h>
#include <stdint.h>

struct intc_group
{
    uint32_t itr;
    uint32_t mir;
    uint32_t mir_clear;
    uint32_t mir_set;
    uint32_t isr_set;
    uint32_t isr_clear;
    uint32_t pending_irq;
    uint32_t pending_fiq;
};

struct intc
{
    uint32_t revision;
    uint32_t padding0[3];
    uint32_t sysconfig;
    uint32_t sysstatus;
    uint32_t padding1[10];
    uint32_t sir_irq;
    uint32_t sir_fiq;
    uint32_t control;
    uint32_t protection;
    uint32_t idle;
    uint32_t padding2[3];
    uint32_t irq_priority;
    uint32_t fiq_priority;
    uint32_t threshold;
    uint32_t padding3[5];
    struct intc_group group[4];
    uint32_t ilr[INTC_NUM_INTERRUPTS];
};

#define INTC ((volatile struct intc *)INTC_BASE)

rt_static_assert(offsetof(struct intc, control) == INTC_CONTROL_OFFSET,
                 "the control field offset in intc is incorrect");
rt_static_assert(offsetof(struct intc, threshold) == INTC_THRESHOLD_OFFSET,
                 "the threshold field offset in intc is incorrect");
rt_static_assert((offsetof(struct intc, group) +
                  offsetof(struct intc_group, isr_clear)) ==
                     INTC_ISR_CLEAR_OFFSET(0),
                 "the isr_clear field offset in intc is incorrect");

extern void (*rt_intc_vector[INTC_NUM_INTERRUPTS])(void);

#define INTC_SYSCONFIG_AUTOIDLE (UINT32_C(1) << 0)
#define INTC_SYSCONFIG_SOFTRESET (UINT32_C(1) << 1)

#define INTC_SYSSTATUS_RESETDONE (UINT32_C(1) << 0)

#define INTC_PROTECTION_ENABLE (UINT32_C(1) << 0)

#define INTC_IDLE_FUNCIDLE (UINT32_C(1) << 0)
#define INTC_IDLE_TURBO (UINT32_C(1) << 1)

#define INTC_ILR_PRIORITY(priority) ((uint32_t)(priority) << 2)
#define INTC_ILR_PRIORITY_MASK INTC_ILR_PRIORITY(INTC_PRIORITY_MIN)
#define INTC_ILR_FIQNIRQ (UINT32_C(1) << 0)

static inline void intc_reset(void)
{
    INTC->sysconfig = INTC_SYSCONFIG_SOFTRESET;
    while ((INTC->sysstatus & INTC_SYSSTATUS_RESETDONE) == 0)
    {
        // Wait for INTC reset to complete.
    }
}

static inline void intc_protect(void)
{
    INTC->protection = INTC_PROTECTION_ENABLE;
}

#define INTC_GROUP_REG_SET(reg, i)                                             \
    do                                                                         \
    {                                                                          \
        INTC->group[(i) / 32].reg = UINT32_C(1) << ((i) % 32);                 \
    } while (0)

static inline void intc_enable(uint32_t i)
{
    INTC_GROUP_REG_SET(mir_clear, i);
}

static inline void intc_disable(uint32_t i)
{
    INTC_GROUP_REG_SET(mir_set, i);
}

static inline void intc_set(uint32_t i)
{
    INTC_GROUP_REG_SET(isr_set, i);
}

static inline void intc_clear(uint32_t i)
{
    INTC_GROUP_REG_SET(isr_clear, i);
}

static inline void intc_set_priority(uint32_t i, uint8_t priority)
{
    const uint32_t ilr = INTC->ilr[i] & ~INTC_ILR_PRIORITY_MASK;
    INTC->ilr[i] = ilr | INTC_ILR_PRIORITY(priority);
}

enum intc_type
{
    INTC_TYPE_IRQ,
    INTC_TYPE_FIQ,
};

static inline void intc_set_type(uint32_t i, enum intc_type type)
{
    if (type == INTC_TYPE_FIQ)
    {
        INTC->ilr[i] |= INTC_ILR_FIQNIRQ;
    }
    else
    {
        INTC->ilr[i] &= ~INTC_ILR_FIQNIRQ;
    }
}

#endif // !defined(__ASSEMBLER__)
