#pragma once

@ vim:ft=arm

/* This file provides the entry and exit sequences for the syscall handler
 * specific to the Interrupt Controller (INTC) as implemented on TI Sitara
 * AM335x processors. */

#include "intc.h"
#include "vic.h"

#include <rt/arch/mode.h>
#include <rt/arch/pseudo.S>

.macro vic_svc_start
    /* Set the interrupt priority threshold and re-evaluate IRQs. Re-evaluating
     * is necessary because an FIQ can occur and trigger the SW IRQ while this
     * code runs, which will cause the SW IRQ run again once IRQs are
     * re-enabled, even though its priority is equal to the threshold. */
    mov32 r12, INTC_BASE
    mov lr, INTC_PRIORITY_MIN
    str lr, [r12, INTC_THRESHOLD_OFFSET]
    mov lr, INTC_CONTROL_NEWIRQAGR
    str lr, [r12, INTC_CONTROL_OFFSET]
    /* Force any pending syscall IRQ to be executed immediately when interrupts
     * are re-enabled, rather than after the syscall handler may have called
     * rt_syscall_run already. */
    dsb
    isb
.endm

.macro vic_svc_finish
    mov32 r0, INTC_BASE
    mov r1, INTC_THRESHOLD_DISABLED
    str r1, [r0, INTC_THRESHOLD_OFFSET]
.endm

// TODO: support interrupt nesting.

// TODO: save and restore VFP context if required.

    .section .text.rt_irq_handler,"ax",%progbits
    .global rt_irq_handler
    .type rt_irq_handler, %function
rt_irq_handler:
    sub lr, 4
    srsdb sp!, MODE_SYS
    cps MODE_SYS
    push {r0-r3, r12, lr}
    cps MODE_IRQ
    // NOTE: r1 is used by rt_syscall_handler.
    mov32 r1, INTC_BASE
    mov32 r2, rt_intc_vector
    ldr r0, [r1, INTC_SIR_IRQ_OFFSET]
    and r0, INTC_SIR_ACTIVE_MASK
    ldr r2, [r2, r0, lsl 2]
    blx r2
    mov32 r1, INTC_BASE
    mov r2, INTC_CONTROL_NEWIRQAGR
    str r2, [r1, INTC_CONTROL_OFFSET]
    cps MODE_SYS
    pop {r0-r3, r12, lr}
    rfeia sp!

    .size rt_irq_handler, .-rt_irq_handler


    /* The FIQ sequence is nearly the same as the IRQ sequence, except that
     * r8-r11 can be used without saving them first, but this only helps for
     * INTC_BASE. r12 doesn't need to be saved and restored either, but adding
     * it preserves stack alignment. */
    .section .text.rt_fiq_handler,"ax",%progbits
    .global rt_fiq_handler
    .type rt_fiq_handler, %function
rt_fiq_handler:
    push {r0-r3, r12, lr}
    mov32 r8, INTC_BASE
    mov32 r2, rt_intc_vector
    ldr r0, [r8, INTC_SIR_FIQ_OFFSET]
    and r0, INTC_SIR_ACTIVE_MASK
    ldr r2, [r2, r0, lsl 2]
    blx r2
    mov r2, INTC_CONTROL_NEWIRQAGR | INTC_CONTROL_NEWFIQAGR
    str r2, [r8, INTC_CONTROL_OFFSET]
    pop {r0-r3, r12, lr}
    subs pc, lr, 4

    .size rt_fiq_handler, .-rt_fiq_handler


    .section .text.rt_syscall_handler,"ax",%progbits
    .global rt_syscall_handler
    .type rt_syscall_handler, %function
rt_syscall_handler:
    str lr, [sp, #-8]!
    // Clear the SW IRQ.
    // r1 still contains INTC_BASE.
    mov r0, (1U << (RT_INTC_SW_IRQ % 32))
    str r0, [r1, INTC_ISR_CLEAR_OFFSET(RT_INTC_SW_IRQ / 32)]
    bl rt_syscall_run_pending
    cbz r0, 0f
    cps MODE_SYS
    swapcontext
    cps MODE_IRQ
0:  ldr pc, [sp], 8


// INTC doesn't support vectored mode, so prevent rt_syscall_irq_handler from
// being used.
.macro vic_syscall_irq_start
    bkpt
.endm

.macro vic_syscall_irq_finish
    bkpt
.endm
