@ vim:ft=arm

/* The interrupt management code for the syscall handler depends on both the
 * type of interrupt controller and the source of the syscall interrupt, so we
 * provide a way to extend this code with assembly macros.
 * vic_syscall_start needs to mask and clear the syscall interrupt such that
 * another syscall cannot occur when IRQs are re-enabled. vic_syscall_finish
 * must unmask the syscall interrupt. Both are run with IRQs disabled. */

#include <rt/mpu.h>

#include <rt/arch/fp.h>
#include <rt/arch/mode.h>
#include <rt/arch/profile.h>
#include <rt/arch/coprocessor.h>
#include <rt/arch/pseudo.S>

    .syntax unified

#ifdef __thumb__
    .thumb
#else // arm
    .arm
#endif // __thumb__

#if RT_ARM_FP
#if RT_ARM_FP_D32
#define pushfp vpush {d16-d31}; vpush {d0-d15}
#define popfp vpop {d0-d15}; vpop {d16-d31}
#else // !RT_ARM_FP_D32
#define pushfp vpush {d0-d15}
#define popfp vpop {d0-d15}
#endif // RT_ARM_FP_D32
#endif // RT_ARM_FP

.macro mpuconfigure
#if RT_MPU_TASK_REGIONS_ENABLE
    mov32 r1, rt_active_task
    ldr r1, [r1]
    adds r1, RT_MPU_TASK_CONFIG_OFFSET
    movs r2, RT_MPU_TASK_REGION_START_ID
    adds r5, r2, RT_MPU_NUM_TASK_REGIONS

.Lmpusetloop\@:
    // Set the region number register.
    mcr p15, 0, r2, c6, c2, 0

    // Load the region configuration and advance to the next one.
    ldmia r1!, {r3, r4}

    // Set the base address register.
    mcr p15, 0, r3, c6, c1, 0

    // Set the size and enable register.
    uxtb r3, r4
    mcr p15, 0, r3, c6, c1, 2

    // Set the access control register.
    lsrs r3, r4, 16
    mcr p15, 0, r3, c6, c1, 4

    add r2, 1
    cmp r2, r5
    blt .Lmpusetloop\@
    dsb
#endif
.endm

.macro swapcontext
#if RT_ARM_FP
    // Check if the task has a floating-point context.
    mrc p15, 0, r1, cr1, cr0, 2
    tst r1, CPACR_FPU_ENABLE
    beq .Lskip_fp_save\@
    pushfp
    vmrs r2, fpexc
    vmrs r3, fpscr
    push {r2, r3}
.Lskip_fp_save\@:
    push {r1, r4-r11} // cpacr, callee-saved registers
#else // !RT_ARM_FP
    push {r4-r11}
#endif // RT_ARM_FP

#if RT_MPU_TASK_REGIONS_ENABLE
/* Force an exception if the context was just saved past the end of the task's
 * stack. This may unrecoverable, as the stores have already occurred into a
 * disallowed area, unless this is specifically set up as a stack guard. */
    ldrt r4, [sp]
#endif // RT_MPU_TASK_REGIONS_ENABLE

    // Store the stack pointer with the saved context.
    mov32 r2, rt_context_prev
    ldr r2, [r2]
    str sp, [r2]

    mpuconfigure

    // Switch to the new task stack.
    mov sp, r0

#if RT_ARM_FP
    pop {r1, r4-r11}
    mcr p15, 0, r1, cr1, cr0, 2
    tst r1, CPACR_FPU_ENABLE
    beq .Lskip_fp_restore\@
    pop {r2, r3}
    vmsr fpexc, r2
    vmsr fpscr, r3
    popfp
.Lskip_fp_restore\@:
#else // !RT_ARM_FP
    pop {r4-r11}
#endif // RT_ARM_FP
.endm


#include <vic.S>


    .section .text.rt_svc_handler,"ax",%progbits
    .global rt_svc_handler
    .type rt_svc_handler, %function
rt_svc_handler:
    srsdb sp!, MODE_SYS
    /* The task passed syscall arguments in r0-r3 and expects them to be
     * clobbered, along with r12 and lr, so none of them need to be saved, and
     * r12 and lr can be used in vic_svc_start. */
    vic_svc_start
    cpsie i
    bl rt_syscall_run
    cbnz r0, 0f
    /* If no context switch is required, we can resume the current task without
     * restoring volatile registers. */
    cpsid i, MODE_SYS
    vic_svc_finish
    rfeia sp!
0:  cps MODE_SYS
    // Reserve space for volatile registers, but no need to save them.
    sub sp, 24
    /* Holding an exclusive monitor across a system call is technically
     * possible, but rt_syscall's svc inline assembly has a memory clobber on
     * it, so this shouldn't happen in real code. Therefore, don't bother
     * clearing the reservation. */
    swapcontext
    cpsid i
    vic_svc_finish
    pop {r0-r3, r12, lr}
    rfeia sp!

    .size rt_svc_handler, .-rt_svc_handler


    .section .text.rt_syscall_irq_handler,"ax",%progbits
    .global rt_syscall_irq_handler
    .type rt_syscall_irq_handler, %function
#ifdef RT_VIC_IRQ_ALIGN
    .balign RT_VIC_IRQ_ALIGN
#endif // RT_VIC_IRQ_ALIGN
rt_syscall_irq_handler:
    sub lr, 4
    srsdb sp!, MODE_SYS
    cps MODE_SYS
    push {r0-r3, r12, lr}
    vic_syscall_irq_start
    cpsie i, MODE_SVC
    bl rt_syscall_run_pending
    cps MODE_SYS
    cbz r0, 0f
    /* rt_syscall_run_pending always uses an atomic instruction, so the
     * exclusive reservation will already be cleared at this point. */
    swapcontext
0:  cpsid i
    vic_syscall_irq_finish
    pop {r0-r3, r12, lr}
    rfeia sp!

    .size rt_syscall_irq_handler, .-rt_syscall_irq_handler


    .section .text.rt_start,"ax",%progbits
    .global rt_start
    .type rt_start, %function
rt_start:
    bl rt_start_context
    cps MODE_SYS
    mpuconfigure
    mov sp, r0

#if RT_ARM_FP
    pop {r1, r4-r11}
    mcr p15, 0, r1, cr1, cr0, 2
#else // !RT_ARM_FP
    pop {r4-r11}
#endif // RT_ARM_FP

    pop {r0-r3, r12, lr}
    rfeia sp!

    .size rt_start, .-rt_start
