#pragma once

#include <rt/arch/coprocessor.h>
#include <rt/arch/fp.h>
#include <rt/arch/profile.h>

#include <vic.h>

#if RT_ARM_FP

#if RT_ARM_FP_D32
#define RT_IRQ_PUSH_FP __asm__("vpush {d16-d31}; vpush {d0-d7}" ::: "memory")
#define RT_IRQ_POP_FP __asm__("vpop {d0-d7}; vpop {d16-d31}" ::: "memory")
#else // !RT_ARM_FP_D32
#define RT_IRQ_PUSH_FP __asm__("vpush {d0-d7}" ::: "memory")
#define RT_IRQ_POP_FP __asm__("vpop {d0-d7}" ::: "memory")
#endif // RT_ARM_FP_D32

/* Define an interrupt handler that automatically saves and restores
 * floating-point state, including CPACR, which tasks use to track whether the
 * floating-point context is active. The handler body is put into a separate
 * noinline function to avoid any stack accesses it does from being invalidated
 * by the asm vpush/vpop, which the compiler isn't aware of. Use
 * general-regs-only to prevent the compiler from doing any floating-point
 * state manipulation other than the inline assembly statements. */

#define RT_HANDLER_FP(fn, interrupt_type)                                      \
    static void fn##_body(void);                                               \
    _Pragma("GCC diagnostic push");                                            \
    _Pragma("GCC diagnostic ignored \"-Wattributes\"");                        \
    __attribute__((target("general-regs-only"), interrupt(interrupt_type),     \
                   aligned(RT_VIC_IRQ_ALIGN))) void                            \
    fn(void)                                                                   \
    {                                                                          \
        uint32_t old_cpacr = cpacr();                                          \
        cpacr_set(old_cpacr | CPACR_FPU_ENABLE);                               \
        uint32_t old_fpscr, old_fpexc;                                         \
        __asm__ __volatile__("vmrs %0, fpscr" : "=r"(old_fpscr));              \
        __asm__ __volatile__("vmrs %0, fpexc" : "=r"(old_fpexc));              \
        __asm__("vmsr fpscr, %0" : : "r"(0));                                  \
        __asm__("vmsr fpexc, %0" : : "r"(RT_FPEXC_EN));                        \
        RT_IRQ_PUSH_FP;                                                        \
        fn##_body();                                                           \
        RT_IRQ_POP_FP;                                                         \
        __asm__("vmsr fpscr, %0" : : "r"(old_fpscr));                          \
        __asm__("vmsr fpexc, %0" : : "r"(old_fpexc));                          \
        cpacr_set(old_cpacr);                                                  \
    }                                                                          \
    _Pragma("GCC diagnostic pop");                                             \
    __attribute__((noinline)) static void fn##_body(void)

#define RT_HANDLER_NOFP(fn, interrupt_type)                                    \
    _Pragma("GCC diagnostic push");                                            \
    _Pragma("GCC diagnostic ignored \"-Wattributes\"");                        \
    __attribute__((target("general-regs-only"), interrupt(interrupt_type),     \
                   aligned(RT_VIC_IRQ_ALIGN))) void                            \
    fn(void);                                                                  \
    _Pragma("GCC diagnostic pop");                                             \
    void fn(void)

#else // !RT_ARM_FP

#define RT_HANDLER_NOFP(fn, interrupt_type)                                    \
    __attribute__((interrupt(interrupt_type))) void fn(void)

// With softfp, no special handling is required.
#define RT_HANDLER_FP RT_HANDLER_NOFP

#endif // RT_ARM_FP

#define RT_IRQ_HANDLER_FP(fn) RT_HANDLER_FP(fn, "IRQ")
#define RT_FIQ_HANDLER_FP(fn) RT_HANDLER_FP(fn, "FIQ")

#define RT_IRQ_HANDLER_NOFP(fn) RT_HANDLER_NOFP(fn, "IRQ")
#define RT_FIQ_HANDLER_NOFP(fn) RT_HANDLER_NOFP(fn, "FIQ")
