@ vim:ft=arm

// This variant of the syscall handler supports armv{7,8}-m{,+nofp}.

#include <rt/mpu.h>

#include <rt/arch/fp.h>
#include <rt/arch/pseudo.S>

#include "exc_return.h"

    .syntax unified

#if RT_ARM_FP
/* For floating-point, save and restore lr, which contains the exception return
 * value, and use the floating-point context bit of this value to decide if the
 * non-volatile fp context should be saved/restored as well. */
#define excreturn ,lr
#define pushlr str lr, [sp, #-8]! // Push only lr but keep sp 8-byte aligned.
#define poplr ldr lr, [sp], 8
#define savefp tst lr, 0x10; it eq; vstmdbeq r1!, {s16-s31}
#define loadfp tst lr, 0x10; it eq; vldmiaeq r0!, {s16-s31}
#define return bx lr
#else // !RT_ARM_FP
/* With nofp, lr can be clobbered and reloaded from an immediate because the
 * exception return value for tasks is a constant. */
#define excreturn
#define pushlr
#define poplr
#define savefp
#define loadfp
#define return mov r0, TASK_INITIAL_EXC_RETURN; bx r0
#endif // RT_ARM_FP

#if RT_MPU_TASK_REGIONS_ENABLE
#define controltemp r2,
#define getcontrol mrs r2, control
#define setcontrol msr control, r2
/* Force an exception if the context was just saved past the end of the task's
 * stack. This may unrecoverable, as the stores have already occurred into a
 * disallowed area, unless this is specifically set up as a stack guard. */
#define teststack ldrt r4, [r1]

.macro mpuconfigure
    mov32 r1, rt_active_task
    ldr r1, [r1]
    adds r1, RT_MPU_TASK_CONFIG_OFFSET
    mov32 r2, RT_MPU_BASE + RT_MPU_REGION_OFFSET
    mpuset RT_MPU_TASK_REGION_START_ID, RT_MPU_NUM_TASK_REGIONS
    dsb
.endm

.macro mpuset r, n
    .if \n == 0
    .exitm
    .endif

#if (__ARM_ARCH == 8) && (RT_MPU_NUM_TASK_REGIONS > 4)
    .if (\r == RT_MPU_TASK_REGION_START_ID) || ((\r % 4) == 0)
    mov r3, \r & 0xFC
    str r3, [r2, -4]
    .endif
#endif

    .if ((\r % 4) == 0) && (\n >= 4)
    ldmia r1!, {r4-r11}
    stmia r2, {r4-r11}
    mpuset (\r + 4), (\n - 4)
    .elseif ((\r % 4) == 1) && (\n >= 3)
    ldmia r1!, {r4-r9}
#if __ARM_ARCH == 8
    add r3, r2, 4
    stmia r3, {r4-r9}
#else // __ARM_ARCH == 7
    stmia r2, {r4-r9}
#endif // __ARM_ARCH
    mpuset (\r + 3), (\n - 3)
    .elseif ((\r % 4) == 2) && (\n >= 2)
    ldmia r1!, {r4-r7}
#if __ARM_ARCH == 8
    add r3, r2, 8
    stmia r3, {r4-r7}
#else // __ARM_ARCH == 7
    stmia r2, {r4-r7}
#endif // __ARM_ARCH
    mpuset (\r + 2), (\n - 2)
    .else
    ldmia r1!, {r4-r5}
#if __ARM_ARCH == 8
    strd r4, r5, [r2, (\r % 4) * 4]
#else // __ARM_ARCH == 7
    strd r4, r5, [r2]
#endif // __ARM_ARCH
    mpuset (\r + 1), (\n - 1)
    .endif
.endm

#else // !RT_MPU_TASK_REGIONS_ENABLE
#define controltemp
#define getcontrol
#define setcontrol
#define mpuconfigure
#define teststack
#endif // RT_MPU_TASK_REGIONS_ENABLE

#if __ARM_ARCH == 8
#define psplimtemp r3,
#define getpsplim mrs r3, psplim
#define setpsplim msr psplim, r3
#else // __ARM_ARCH == 7
#define psplimtemp
#define getpsplim
#define setpsplim
#endif // __ARM_ARCH

#define saveregs stmdb r1!, {controltemp psplimtemp r4-r11 excreturn}
#define loadregs ldmia r0!, {controltemp psplimtemp r4-r11 excreturn}

.macro swapcontext
    // Write the suspending context to its stack pointer.
    mrs r1, psp
    savefp
    getcontrol
    getpsplim
    saveregs
    teststack

    // Store the new stack pointer with the saved context.
    mov32 r2, rt_context_prev
    ldr r2, [r2]
    str r1, [r2]

    mpuconfigure

    // Load the new context from r0.
    loadregs
    loadfp

    // Set the new stack pointer.
    msr psp, r0
    setcontrol
    setpsplim
.endm

    .section .text.rt_svcall_handler,"ax",%progbits
    .global rt_svcall_handler
    .type rt_svcall_handler, %function
rt_svcall_handler:
    pushlr
    bl rt_syscall_run
    poplr
    cbz r0, 0f
    swapcontext
0:  return

    .size rt_svcall_handler, .-rt_svcall_handler


    .section .text.rt_pendsv_handler,"ax",%progbits
    .global rt_pendsv_handler
    .type rt_pendsv_handler, %function
rt_pendsv_handler:
    pushlr
    bl rt_syscall_run_pending
    poplr
    cbz r0, 0f
    swapcontext
0:  return

    .size rt_pendsv_handler, .-rt_pendsv_handler


    .section .text.rt_start,"ax",%progbits
    .global rt_start
    .type rt_start, %function
rt_start:
    bl rt_start_context
    mpuconfigure
    loadregs
    msr psp, r0
    setpsplim
    // Set the SPSEL bit in control to switch to the process stack pointer.
    movs r0, 2
    msr control, r0
    isb
    // Once we are using the process stack pointer, interrupts can be enabled.
    cpsie i
    /* Manually perform an exception return stack pop. This ignores the
     * context's initial PSR, but the current PSR works as well. */
    pop {r0-r3, r12, lr}
    ldr pc, [sp], 8

    .size rt_start, .-rt_start
