use core::{
    cell::UnsafeCell,
    ffi::{c_uchar, c_void},
    marker::{PhantomData, PhantomPinned},
    mem::{size_of, MaybeUninit},
    panic::{RefUnwindSafe, UnwindSafe},
    ptr::from_ref,
};

use crate::{
    bindings::{
        rt_queue, rt_queue_peek, rt_queue_pop, rt_queue_push, rt_queue_timedpeek,
        rt_queue_timedpop, rt_queue_timedpush, rt_queue_trypeek, rt_queue_trypop, rt_queue_trypush,
    },
    sync::sem::c_sem_init,
    tick::Utick,
};

/// A multi-producer, multi-consumer, bounded queue that supports blocking, non-blocking, and timed
/// push, pop, and peek operations. The queue is lock-free for up to `n` tasks where `n` is the
/// depth of the queue.
#[repr(transparent)]
pub struct Queue<T: ?Sized> {
    q: UnsafeCell<rt_queue>,
    _phantom_elem: PhantomData<T>,
    _pin_marker: PhantomPinned,
}

unsafe impl<T: ?Sized + Send> Send for Queue<T> {}
unsafe impl<T: ?Sized + Send> Sync for Queue<T> {}
impl<T: ?Sized> UnwindSafe for Queue<T> {}
impl<T: ?Sized> RefUnwindSafe for Queue<T> {}

impl<T> Queue<T> {
    pub const fn init(
        queue: &'static Queue<T>,
        slots: *const c_uchar,
        data: *const c_void,
        num_elems: usize,
    ) -> Queue<T> {
        let q = unsafe { &*queue.q.get().cast_const() };
        Queue {
            q: UnsafeCell::new(rt_queue {
                push_sem: c_sem_init(&q.push_sem, num_elems as i32, i32::max_value()),
                pop_sem: c_sem_init(&q.pop_sem, 0, i32::max_value()),
                enq: 0,
                deq: 0,
                slots: slots.cast_mut(),
                data: data.cast_mut(),
                num_elems,
                elem_size: size_of::<T>(),
            }),
            _phantom_elem: PhantomData,
            _pin_marker: PhantomPinned,
        }
    }

    /// Adds an element to the queue, blocking until there is space in the queue.
    pub fn push(&self, elem: T) {
        unsafe { rt_queue_push(self.q.get(), from_ref(&elem).cast()) }
    }

    /// Removes an element from the queue, blocking while the queue is empty.
    pub fn pop(&self) -> T {
        let mut elem = MaybeUninit::<T>::uninit();
        unsafe {
            rt_queue_pop(self.q.get(), elem.as_mut_ptr().cast());
            elem.assume_init()
        }
    }

    /// Gets the next element of the queue, blocking while the queue is empty.
    pub fn peek(&self) -> T {
        let mut elem = MaybeUninit::<T>::uninit();
        unsafe {
            rt_queue_peek(self.q.get(), elem.as_mut_ptr().cast());
            elem.assume_init()
        }
    }

    /// Adds an element to the queue if there is space and returns the element to the caller
    /// otherwise.
    pub fn try_push(&self, elem: T) -> Result<(), T> {
        if unsafe { rt_queue_trypush(self.q.get(), from_ref(&elem).cast()) } {
            Ok(())
        } else {
            Err(elem)
        }
    }

    /// Removes an element from the queue if it is non-empty.
    pub fn try_pop(&self) -> Option<T> {
        let mut elem = MaybeUninit::<T>::uninit();
        if unsafe { rt_queue_trypop(self.q.get(), elem.as_mut_ptr().cast()) } {
            Some(unsafe { elem.assume_init() })
        } else {
            None
        }
    }

    /// Gets the next element of the queue if it is non-empty.
    pub fn try_peek(&self) -> Option<T> {
        let mut elem = MaybeUninit::<T>::uninit();
        if unsafe { rt_queue_trypeek(self.q.get(), elem.as_mut_ptr().cast()) } {
            Some(unsafe { elem.assume_init() })
        } else {
            None
        }
    }

    pub fn timed_push(&self, elem: T, ticks: Utick) -> Result<(), T> {
        if unsafe { rt_queue_timedpush(self.q.get(), from_ref(&elem).cast(), ticks) } {
            Ok(())
        } else {
            Err(elem)
        }
    }

    pub fn timed_pop(&self, ticks: Utick) -> Option<T> {
        let mut elem = MaybeUninit::<T>::uninit();
        if unsafe { rt_queue_timedpop(self.q.get(), elem.as_mut_ptr().cast(), ticks) } {
            Some(unsafe { elem.assume_init() })
        } else {
            None
        }
    }

    pub fn timed_peek(&self, ticks: Utick) -> Option<T> {
        let mut elem = MaybeUninit::<T>::uninit();
        if unsafe { rt_queue_timedpeek(self.q.get(), elem.as_mut_ptr().cast(), ticks) } {
            Some(unsafe { elem.assume_init() })
        } else {
            None
        }
    }
}

#[macro_export]
macro_rules! queue {
    ($name: ident, $type: ty, $num: literal) => {
        $crate::paste::paste! {
            static mut [< $name _SLOTS >]: [u8; $num] = [0u8; $num];
            static mut [< $name _DATA >]: [$type; $num] = unsafe { core::mem::zeroed() };
            static $name: $crate::sync::Queue<$type> = $crate::sync::Queue::init(
                &$name,
                unsafe { core::ptr::addr_of!([< $name _SLOTS >][0]) },
                unsafe { core::ptr::addr_of!([< $name _DATA >][0]).cast() },
                $num
            );
        }
    };
}

#[cfg(test)]
mod tests {
    queue!(QUEUE, i32, 10);

    #[test]
    fn fast_path() {
        QUEUE.push(1);
        assert!(QUEUE.pop() == 1);
        assert!(QUEUE.try_pop().is_none());
    }
}
